@@ -13,10 +13,12 @@ import json
 import io
 import inspect
 import itertools
+import logging
 import mimetypes
 import os
 import random
 import ssl
+import socket
 import time
 import uuid
 import urllib.parse
@@ -43,7 +45,9 @@ def request(method, url, *,
             expect100=False,
             session=None,
             verify_ssl=True,
-            loop=None):
+            connection_params=None,
+            loop=None,
+            request_class=None):
     """Constructs and sends a request. Returns response object.
 
     method: http method
@@ -84,18 +88,20 @@ def request(method, url, *,
     redirects = 0
     if loop is None:
         loop = asyncio.get_event_loop()
+    if request_class is None:
+        request_class = HttpRequest
 
     while True:
-        req = HttpRequest(
+        req = request_class(
             method, url, params=params, headers=headers, data=data,
             cookies=cookies, files=files, auth=auth, encoding=encoding,
             version=version, compress=compress, chunked=chunked,
             verify_ssl=verify_ssl, loop=loop, expect100=expect100)
 
         if session is None:
-            conn = _connect(req, loop)
+            conn = _connect(req, loop, connection_params)
         else:
-            conn = session.start(req, loop)
+            conn = session.start(req, loop, connection_params)
 
         if conn_timeout is None and timeout is not None:
             conn_timeout = timeout
@@ -112,8 +118,10 @@ def request(method, url, *,
                 transport, proto, req, wrp, timeout, loop)
         except asyncio.TimeoutError:
             raise aiohttp.TimeoutError from None
+        except aiohttp.BadStatusLine as exc:
+            raise aiohttp.ClientConnectionError(exc)
         except OSError as exc:
-            raise aiohttp.ConnectionError(exc)
+            raise aiohttp.OsConnectionError(exc)
         finally:
             conn_task.cancel()
 
@@ -143,10 +151,17 @@ def request(method, url, *,
 
 
 @asyncio.coroutine
-def _connect(req, loop):
-    transport, proto = yield from loop.create_connection(
-        functools.partial(aiohttp.StreamProtocol, loop=loop),
-        req.host, req.port, ssl=req.ssl)
+def _connect(req, loop, params):
+    if params is not None:
+        transport, proto = yield from loop.create_connection(
+            functools.partial(aiohttp.StreamProtocol, loop=loop),
+            params['host'], params['port'],
+            ssl=params['ssl'], family=params['family'],
+            proto=params['proto'], flags=params['flags'])
+    else:
+        transport, proto = yield from loop.create_connection(
+            functools.partial(aiohttp.StreamProtocol, loop=loop),
+            req.host, req.port, ssl=req.ssl)
     wrp = TransportWrapper(transport)
     return transport, proto, wrp
 
@@ -181,10 +196,12 @@ class HttpClient:
     mark failed hosts.
     """
 
+    _resolve_timeout = 360.0  # update dns info every 5 minutes
+
     def __init__(self, hosts, *,
-                 method=None, path=None, ssl=False,
+                 method=None, path=None, ssl=False, session=False,
                  timeout=None, conn_timeout=None, failed_timeout=5.0,
-                 loop=None):
+                 resolve=True, loop=None):
         super().__init__()
 
         if isinstance(hosts, str):
@@ -215,7 +232,7 @@ class HttpClient:
         self._timeout = timeout
         self._conn_timeout = conn_timeout
         self._schema = 'https' if ssl else 'http'
-        self._session = aiohttp.Session()
+        self._session = aiohttp.Session() if session else None
 
         self._failed = collections.deque()
         self._failed_handle = None
@@ -225,6 +242,38 @@ class HttpClient:
             loop = asyncio.get_event_loop()
         self._loop = loop
 
+        self._resolve = resolve
+        self._resolved_hosts = {}
+        if resolve:
+            self._resolve_handle = self._loop.call_later(
+                self._resolve_timeout, self._cleanup_resolved_host)
+
+    def _cleanup_resolved_host(self):
+        self._resolved_hosts.clear()
+        self._resolve_handle = self._loop.call_later(
+            self._resolve_timeout, self._cleanup_resolved_host)
+
+    @asyncio.coroutine
+    def _resolve_host(self, host, port):
+        if self._resolve:
+            key = (host, port)
+            if key not in self._resolved_hosts:
+                infos = yield from self._loop.getaddrinfo(
+                    host, port, type=socket.SOCK_STREAM, family=socket.AF_INET)
+
+                hosts = []
+                for family, _, proto, _, address in infos:
+                    hosts.append(
+                        {'host': address[0], 'port': address[1],
+                         'ssl': self._ssl, 'family': family,
+                         'proto': proto, 'flags': socket.AI_NUMERICHOST})
+                self._resolved_hosts[key] = hosts
+
+            return self._resolved_hosts[key]
+        else:
+            return [{'host': host, 'port': port,
+                     'ssl': self._ssl, 'family': 0, 'proto': 0, 'flags': 0}]
+
     def _resurrect_failed(self):
         now = int(time.time())
 
@@ -277,29 +326,40 @@ class HttpClient:
         while hosts:
             idx = random.randint(0, len(hosts)-1)
 
-            host_info = hosts[idx]
+            h_info = hosts[idx]
             url = urllib.parse.urljoin(
                 '{}://{}:{}'.format(
-                    self._schema, host_info[0], host_info[1]), path)
-            try:
-                resp = yield from request(
-                    method, url, params=params, data=data, headers=headers,
-                    cookies=cookies, files=files, auth=auth, encoding=encoding,
-                    allow_redirects=allow_redirects, version=version,
-                    max_redirects=max_redirects, conn_timeout=conn_timeout,
-                    timeout=timeout, compress=compress, chunked=chunked,
-                    verify_ssl=verify_ssl, expect100=expect100,
-                    session=self._session, loop=self._loop)
-            except (aiohttp.ConnectionError, aiohttp.TimeoutError):
-                if host_info in hosts:
-                    # could be removed concurrently
-                    hosts.remove(host_info)
-                    self._failed.append((host_info, int(time.time())))
-                    if not self._failed_handle:
-                        self._failed_handle = self._loop.call_later(
-                            self._failed_timeout, self._resurrect_failed)
-            else:
-                return resp
+                    self._schema, h_info[0], h_info[1]), path)
+            host_params = yield from self._resolve_host(h_info[0], h_info[1])
+
+            for conn_params in host_params:
+                try:
+                    resp = yield from request(
+                        method, url, params=params, data=data, headers=headers,
+                        cookies=cookies, files=files, auth=auth,
+                        encoding=encoding, allow_redirects=allow_redirects,
+                        version=version, max_redirects=max_redirects,
+                        conn_timeout=conn_timeout, timeout=timeout,
+                        compress=compress, chunked=chunked,
+                        verify_ssl=verify_ssl, expect100=expect100,
+                        session=self._session, connection_params=conn_params,
+                        loop=self._loop)
+                except (aiohttp.ConnectionError, aiohttp.TimeoutError):
+                    pass
+                else:
+                    return resp
+
+            if h_info in hosts:
+                # could be removed concurrently
+                hosts.remove(h_info)
+                self._failed.append((h_info, int(time.time())))
+                if not self._failed_handle:
+                    self._failed_handle = self._loop.call_later(
+                        self._failed_timeout, self._resurrect_failed)
+
+                key = (h_info[0], h_info[1])
+                if key in self._resolved_hosts:
+                    del self._resolved_hosts[key]
 
         raise aiohttp.ConnectionError('All hosts are unreachable.')
 
@@ -319,7 +379,7 @@ class HttpRequest:
     auth = None
     response = None
 
-    _writer = None  # async task for streaming body
+    _writer = None  # async task for streaming data
     _continue = None  # waiter future for '100 Continue' response
 
     def __init__(self, method, url, *,
@@ -619,24 +679,37 @@ class HttpRequest:
             yield from self._continue
 
         if is_stream:
+            exc = None
             value = None
             stream = self.body
 
             while True:
                 try:
-                    result = stream.send(value)
+                    if exc is not None:
+                        result = stream.throw(exc)
+                    else:
+                        result = stream.send(value)
                 except StopIteration as exc:
                     if isinstance(exc.value, bytes):
                         request.write(exc.value)
                     break
+                except:
+                    self.response.close(True)
+                    raise
 
                 if isinstance(result, asyncio.Future):
-                    value = yield from result
+                    exc = None
+                    value = None
+                    try:
+                        value = yield result
+                    except Exception as err:
+                        exc = err
                 elif isinstance(result, bytes):
                     request.write(result)
                     value = None
                 else:
-                    raise ValueError('Bytes object is expected.')
+                    raise ValueError(
+                        'Bytes object is expected, got: %s.' % type(result))
         else:
             if isinstance(self.body, bytes):
                 self.body = (self.body,)
@@ -675,7 +748,8 @@ class HttpRequest:
             request.write_eof()
 
         self.response = HttpResponse(
-            self.method, self.path, self.host, self._continue)
+            self.method, self.path, self.host,
+            writer=self._writer, continue100=self._continue)
         return self.response
 
     @asyncio.coroutine
@@ -704,17 +778,20 @@ class HttpResponse(http.client.HTTPMessage):
 
     _response_parser = aiohttp.HttpResponseParser()
 
-    def __init__(self, method, url, host='', continue100=None):
+    def __init__(self, method, url, host='', *, writer=None, continue100=None):
         super().__init__()
 
         self.method = method
         self.url = url
         self.host = host
         self._content = None
+        self._writer = writer
         self._continue = continue100
 
     def __del__(self):
-        self.close()
+        if self.transport is not None:
+            logging.warn('HttpResponse has to be closed explicitly!')
+            self.close()
 
     def __repr__(self):
         out = io.StringIO()
@@ -723,6 +800,8 @@ class HttpResponse(http.client.HTTPMessage):
         print(super().__str__(), file=out)
         return out.getvalue()
 
+    __str__ = __repr__
+
     def start(self, stream, transport):
         """Start response processing."""
         self.stream = stream
@@ -765,6 +844,18 @@ class HttpResponse(http.client.HTTPMessage):
         if self.transport is not None:
             self.transport.close(force)
             self.transport = None
+        if (self._writer is not None) and not self._writer.done():
+            self._writer.cancel()
+            self._writer = None
+
+    @asyncio.coroutine
+    def wait_for_close(self):
+        if self._writer is not None:
+            try:
+                yield from self._writer
+            finally:
+                self._writer = None
+        self.close()
 
     @asyncio.coroutine
     def read(self, decode=False):
